###################################
# Doorbell with face recognition  #
###################################

#########################
# MODULES               #
#########################

import time
import face_recognition
from picamera import PiCamera
import numpy as np
import csv
from PIL import Image
import telegram
from credentials_telegram import token_value
import glob
from practicum import find_mcu_boards, McuBoard, PeriBoard

devs = find_mcu_boards()

if len(devs) == 0:
    print("*** No practicum board found.")
    exit(1)

mcu = McuBoard(devs[0])
print("*** Practicum board found")
print("*** Manufacturer: %s" % \
        mcu.handle.getString(mcu.device.iManufacturer, 256))
print("*** Product: %s" % \
        mcu.handle.getString(mcu.device.iProduct, 256))
peri = PeriBoard(mcu)


COUNTER = 0                 # Counting how much images are obtained (mainly for debugging)
IMG_WIDTH = 640             # Image width
IMG_HEIGHT = 480            # Image height
T_SHOW_RESULT = 10          # Time that the sign is on after the doorbell is rung (in seconds)
CAMERA = PiCamera()         # Instantiate RPi camera object
CAMERA.resolution = (IMG_WIDTH, IMG_HEIGHT)

#########################
# FUNCTIONS             #
#########################

# FACE RECOGNITION FUNCTIONS
# ---------------------------

def get_image_test(fn):
    return np.array(Image.open("img/whitelist/{}.jpg".format(fn)))


def get_face_locations(inp_img):
    return face_recognition.face_locations(inp_img)


def get_face_encoding(inp_img, loc):
    return face_recognition.face_encodings(inp_img, loc)


def match_face_encodings(inp_img, known_faces=None):
    if known_faces == None:
        print('badness at match_face_encodings')

    return face_recognition.compare_faces(known_faces, inp_img)


def save_to_list(enc, list_id):
    print('Save encoding to csv file: {}'.format(list_id))
    with open('{}.csv'.format(list_id), 'a') as f:
        writer = csv.writer(f)
        writer.writerow(enc)


def get_image(filename=None):
    # Debug tool ----
    global COUNTER
    COUNTER = COUNTER + 1
    # print(COUNTER)
    # End of Debug tool ----

    output = np.empty((IMG_HEIGHT, IMG_WIDTH, 3), dtype=np.uint8)
    CAMERA.capture(output, format="rgb")
    if filename is not None:
        print('Saving image: {}.jpg'.format(filename))
        im = Image.fromarray(output)
        im.save("{}.jpeg".format(filename))
    return output


def check_face_from_doorbell():

    # Get whitelist and blacklist from csv file
    whitelist = np.genfromtxt('whitelist.csv', delimiter=',').reshape((-1, 128))
    blacklist = np.genfromtxt('blacklist.csv', delimiter=',').reshape((-1, 128))

    # Take picture and calculate encoding
    enc_doorbell = get_doorbell_enc()

    # Compare encoding with encoding databases
    match_allowed = face_recognition.compare_faces(whitelist, enc_doorbell, tolerance=0.6)
    match_denied  = face_recognition.compare_faces(blacklist, enc_doorbell, tolerance=0.6)

    print('In whitelist: {}\nIn blacklist: {}'.format(any(match_allowed), any(match_denied)))

    # What to do next?
    if any(match_allowed) == True and any(match_denied) == True:
        print('Face encoding in both whitelist and blacklist.')
        output = 2
    elif any(match_allowed) == True:
        print('Face encoding in whitelist')
        output = 1
    elif any(match_denied) == True:
        print('Face encoding in blacklist')
        output = 0
    else:
        print('Face encoding not in whitelist / blacklist')
        output = 3

    return output



def get_doorbell_enc():
    loc = []

    while len(loc) < 1:
        img = get_image(filename='img/doorbell/{}'.format(int(time.time())))
        loc = get_face_locations(img)

    return get_face_encoding(img, loc)


def add_face_to_list(list_id='whitelist'):
    assert (list_id in ['whitelist', 'blacklist']), 'Provide list_id: whitelist/blacklist.'

    # img = get_image_test('face')
    img = get_image(filename='img/whitelist/{}'.format(int(time.time())))
    loc = get_face_locations(img)

    if len(loc) == 1:
        print('Face detected. Calculating encoding...')
        enc = get_face_encoding(img, loc)
        save_to_list(enc[0], list_id)

        peri.set_green(1)
        time.sleep(0.5)
        peri.set_green(0)
        time.sleep(0.5)
        peri.set_green(1)
        time.sleep(0.5)
        peri.set_green(0)
        time.sleep(0.5)
        peri.set_green(1)
        time.sleep(0.5)
        peri.set_green(0)

        print('Done...')
    else:
        if len(loc) > 1:
            print('{} faces detected. Please try again with only 1 face...'.format(len(loc)))
        else:
            print('No face detected. Please try again...'.format(len(loc)))

def add_face_to_list2(list_id='blacklist'):
    assert (list_id in ['whitelist', 'blacklist']), 'Provide list_id: whitelist/blacklist.'

    # img = get_image_test('face')
    img = get_image(filename='img/blacklist/{}'.format(int(time.time())))
    loc = get_face_locations(img)

    if len(loc) == 1:
        print('Face detected. Calculating encoding...')
        enc = get_face_encoding(img, loc)
        save_to_list(enc[0], list_id)

        peri.set_green(1)
        time.sleep(0.5)
        peri.set_green(0)
        time.sleep(0.5)
        peri.set_green(1)
        time.sleep(0.5)
        peri.set_green(0)
        time.sleep(0.5)
        peri.set_green(1)
        time.sleep(0.5)
        peri.set_green(0)

        print('Done...')
    else:
        if len(loc) > 1:
            print('{} faces detected. Please try again with only 1 face...'.format(len(loc)))
        else:
            print('No face detected. Please try again...'.format(len(loc)))



# Telegram Bot
# ---------------------------

def send_img_telegram(msg):
    dir_doorbell = 'img/doorbell/'
    threshold = 60 * 0.5

    fns = sorted([int(val.split('/')[-1].split('.')[0]) for val in glob.glob(dir_doorbell + '*')])

    bot = telegram.Bot(token=token_value)
    chat_id = bot.get_updates()[-1].message.chat_id

    bot.send_message(chat_id=chat_id, text="Doorbell Alarm")
    for fn in fns:
        if fns[-1] - fn < threshold:
            fn_format = dir_doorbell + '{}.jpeg'.format(fn)
            caption_str = msg + ' - {}'.format(time.ctime(fn))
            bot.send_photo(chat_id=chat_id, photo=open(fn_format, 'rb'), caption=caption_str)


#########################
# MAIN                  #
#########################

peri.set_green(0)
peri.set_yellow(0)
peri.set_red(0)
peri.set_blue(0)


while True:

    greensw = peri.get_green()
    yellowsw = peri.get_yellow()
    redsw = peri.get_red()
    bluesw = peri.get_blue()

    if redsw is True:
        """ Red button callback function """
        # Debugging tool:
        print('\nRed button pressed')

        peri.set_red(1)

        # ----------------------------------------------------------------------------------
        

        time.sleep(0.2)

        print('Add face to blacklist.')
        add_face_to_list2()

        # ----------------------------------------------------------------------------------

        peri.set_red(0)

        time.sleep(0.2)

        print('')

    elif yellowsw is True:
        # Debugging tool:
        print('\nYellow button pressed')

        peri.set_yellow(1)

        # ----------------------------------------------------------------------------------
        

        time.sleep(0.2)

        print('Add face to whitelist.')
        add_face_to_list()

        # ----------------------------------------------------------------------------------

        peri.set_yellow(0)

        time.sleep(0.2)

        print('')


    elif greensw is True:
        # Debugging tool:
        print('\nGreen button pressed')

        peri.set_green(1)

        # ----------------------------------------------------------------------------------
        

        time.sleep(2.5)


        # ----------------------------------------------------------------------------------

        peri.set_green(0)

        time.sleep(0.2)

        print('')

    elif bluesw is True:
        print('Bell button pressed')

        peri.set_green(0)
        peri.set_yellow(0)
        peri.set_red(0)
        peri.set_blue(1)
        
        time.sleep(0.5)

        peri.set_blue(0)
        # Check if the face of the person ringing the door exists in database
        # and return result
        result = check_face_from_doorbell()

        # Set the light that represents the result
        # 0 - Encoding on blacklist
        # 1 - Encoding on whitelist
        # 2 - Encoding on whitelist and blacklist (not good)
        # 3 - Encoding not on any of the lists
        if result == 0:
            peri.set_red(1)
            send_img_telegram('\U0001F53A')
            time.sleep(T_SHOW_RESULT)
            peri.set_red(0)
        elif result == 1:
            peri.set_green(1)
            time.sleep(T_SHOW_RESULT)
            peri.set_green(0)
        elif result == 2:
            peri.set_yellow(1)
            send_img_telegram('\U000025FE')
            time.sleep(T_SHOW_RESULT)
            peri.set_yellow(0)
        elif result == 3:
            peri.set_yellow(1)
            send_img_telegram('\U0001F538')
            time.sleep(T_SHOW_RESULT)
            peri.set_yellow(0)

        print('')


